//@ts-check
import * as Y from 'yjs'
import { redo, undo, ySyncPlugin, yUndoPlugin } from 'y-prosemirror'
import { EditorState } from 'prosemirror-state'
import { EditorView } from 'prosemirror-view'
import { keymap } from 'prosemirror-keymap'
import { schema } from './schema'
import getProsemirrorSetup from './prosemirror-setup'
import './styles/main.css'
import { serializeUpdate, deserializeUpdate } from 'webxdc-yjs-provider';
import { WebxdcSyncProvider } from 'webxdc-yjs-provider/initWebxdcSyncProviderGeneric';
import debounce from 'just-debounce-it';
import { getDocName } from './getDocName'
/** @typedef {import('../webxdc.d.ts')} */

let manualSend = localStorage.getItem('manualSend') === 'true';
function toggleManualSend() {
  manualSend = !manualSend;
  localStorage.setItem('manualSend', manualSend.toString());

  updateSyncButtonState();

  // Need to make sure to `sendUnsentLocalUpdates` because otherwise
  // they wouldn't get sent until the page gets closed or a new edit
  // is made to the doc locally.
  if (!manualSend) {
    provider.sendUnsentLocalUpdates();
  }
}
// Currently there is no UI to toggle this.
const alwaysSendOnClose = localStorage.getItem('alwaysSendOnClose') !== 'false';

const ydoc = new Y.Doc()
/** @type {ReturnType<typeof createEditor>} */
let prosemirror
const type = ydoc.getXmlFragment('prosemirror')

let everNotifiedPeersAboutEdits = false;
/** @type {undefined | string} */
let lastSetWebxdcDocumentName = undefined;
/** @type {undefined | string} */
let lastSetWebxdcSummary = undefined;

const provider = new WebxdcSyncProvider(
  ydoc,
  serializeUpdate,
  deserializeUpdate,
  sendUpdate,
);
/**
 * @param {ReturnType<typeof serializeUpdate>} outgoingSerializedYjsUpdate
 */
function sendUpdate(outgoingSerializedYjsUpdate) {
  /** @type {Parameters<typeof webxdc.sendUpdate>[0]} */
  const update = {
    payload: {
      serializedYjsUpdate: outgoingSerializedYjsUpdate,
    },
  };

  const preferredDocumentName = prosemirror && getDocName(prosemirror.state);
  if (
    preferredDocumentName
    && preferredDocumentName !== lastSetWebxdcDocumentName
  ) {
    update.document = preferredDocumentName;
    lastSetWebxdcDocumentName = preferredDocumentName;
  }

  // TODO i18n
  const preferredSummary = `Last edit: ${webxdc.selfName}`;
  if (lastSetWebxdcSummary !== preferredSummary) {
    update.summary = preferredSummary;
  }

  // TODO fix: even if the user undid all of their changes (Ctrl+Z),
  // we still send a message that they "made an edit".
  if (manualSend || !everNotifiedPeersAboutEdits) {
    update.info = formatNotificationAboutEdits(
      preferredDocumentName,
      manualSend,
    );
    everNotifiedPeersAboutEdits = true;
  }

  webxdc.sendUpdate(update, 'Document edited');
  updateSyncButtonState();
}
const storedUpdatesAppliedPromise = webxdc.setUpdateListener(update => {
  if (update.payload?.serializedYjsUpdate) {
    provider.onIncomingYjsUpdate(update.payload.serializedYjsUpdate);
  }
  if (update.document) {
    lastSetWebxdcDocumentName = update.document;
  }
  if (update.summary) {
    lastSetWebxdcSummary = update.summary;
  }
});
{
/**
 * Delta Chat Core throttles updates pretty heavily:
 * https://github.com/deltachat/deltachat-core-rust/blob/b96028cd87f02a83f8f0a5282da4b4bb88cdc05c/src/context.rs#L379
 * https://codeberg.org/webxdc/editor/pulls/23#issuecomment-996164
 * So we better spend the quota wisely, i.e. not spend all of it immediately
 * and then get throttled by the messenger implementation.
 */
const debouncePeriodMs = 5 * 1000;
const debouncedSendLocalUpdatesIfAutosaveEnabled = debounce(
  () => {
    if (manualSend) {
      return;
    }
    provider.sendUnsentLocalUpdates();
  },
  debouncePeriodMs,
);

provider.onNeedToSendLocalUpdates = () => {
  if (!manualSend) {
    debouncedSendLocalUpdatesIfAutosaveEnabled();
  }
  updateSyncButtonState();
}
}

/**
 * @returns {boolean}
 */
function isNeedToSendUnsentUpdatesOnPotentialPageClose() {
  return alwaysSendOnClose || !manualSend;
}

addEventListener("DOMContentLoaded", () => {
  /** @type {HTMLDivElement} */
  const editor = document.querySelector('#editor')
  prosemirror = createEditor(editor);

  // TODO improvement: detach the listener when there are no pending changes?
  // Or is that unnecessary for this event? I heard stuff about
  // back/forward cache.
  window.addEventListener("visibilitychange", () => {
    if (
      document.visibilityState === "hidden" &&
      isNeedToSendUnsentUpdatesOnPotentialPageClose()
    ) {
      provider.sendUnsentLocalUpdates();
    }
  });

  // TODO only add this listener when there are unsaved changes (see
  // the comment right above).
  window.addEventListener(
    "beforeunload",
    maybeSendUpdatesAndDelayCloseUntilTheyreSent
  );

  initExtraButtons();
})

/**
 * @param {string | undefined} docName
 * @param {boolean} manualSend
 */
function formatNotificationAboutEdits(docName, manualSend) {
  // TODO i18n.
  // Could be an empty string, or `undefined`
  const docText = docName != undefined
    ? `'${docName}'`
    : 'the document';
  return `${webxdc.selfName} ${manualSend ? 'edited' : 'editing'} ${docText}`;
}

// A workaround for
// https://github.com/deltachat/deltachat-desktop/issues/3321#issue-1814659377
// https://codeberg.org/webxdc/editor/issues/12#issuecomment-989270
const waitForUpdateSendDelayMs = 100;
let pendingClose = false;
/**
 * @param {BeforeUnloadEvent} e
 */
function maybeSendUpdatesAndDelayCloseUntilTheyreSent(e) {
  if (!isNeedToSendUnsentUpdatesOnPotentialPageClose()) {
    // TODO maybe then prevents page close completely if there are unsaved
    // changes.
    return;
  }

  provider.sendUnsentLocalUpdates();
  if (!pendingClose) {
    pendingClose = true;
    setTimeout(() => {
      window.removeEventListener(
        "beforeunload",
        maybeSendUpdatesAndDelayCloseUntilTheyreSent
      );
      // Yes, we close the parent. This relies on the fact that we run inside an
      // `<iframe>`, which is the case for Delta Chat Desktop.
      try {
        window.parent.close();
      } catch (e) {
        window.close();
      }
    }, waitForUpdateSendDelayMs);
  }

  e.preventDefault();
  // For compatibility with older browsers
  e.returnValue = '';
  return '';
}

/**
 * @param {HTMLElement} element
 */
function createEditor(element) {
  return new EditorView(element, {
    state: EditorState.create({
      schema,
      // TODO refactor perhaps all this belongs to `getProsemirrorSetup`?
      plugins: [
        ySyncPlugin(type),
        // There is also `prosemirror-history` module, but I believe
        // yUndoPlugin is supposed to be a complete replacement for it.
        // See https://codeberg.org/webxdc/editor/issues/13#issuecomment-1006670
        yUndoPlugin(),
        keymap({
          'Mod-z': undo,
          'Mod-y': redo,
          'Mod-Shift-z': redo,
          'Mod-s': () => {
            provider.sendUnsentLocalUpdates();
            return true;
          },
        }),
      ].concat(getProsemirrorSetup({
        schema,
        undo,
        redo,
        getIsManualSend: () => manualSend,
        onToggleManualSendClick: toggleManualSend,
      })),
    }),
  });
}

/** @type {() => void} */
let updateSyncButtonState = () => {};

function initExtraButtons() {
  const extraButtonsWrapper = /** @type {HTMLSpanElement} */
    (document.getElementById('extra-buttons'));
  const menuBar = document.getElementsByClassName('ProseMirror-menubar')[0];
  menuBar.appendChild(extraButtonsWrapper);
  extraButtonsWrapper.style.display = '';

  const syncButton = /** @type {HTMLButtonElement} */
    (document.getElementById('sync'));
  syncButton.addEventListener('click', () => {
    provider.sendUnsentLocalUpdates();
  });
  updateSyncButtonState = () => {
    syncButton.disabled = !provider.haveUnsentUpdates;

    // Actually, let's always show the button, for feedback, and to keep it
    // possible to send the updates when the user wants to.
    // However this makes the user think about the button, and they'll want
    // to click it every time they want to close the app.
    //
    // syncButton.style.display = manualSend ? '' : 'none';
  }
  updateSyncButtonState();
}
