//@ts-check
import { Plugin } from "prosemirror-state"
import { keymap } from "prosemirror-keymap"
import {
  buildInputRules,
  buildKeymap,
  buildMenuItems,
} from "prosemirror-example-setup";
import { baseKeymap } from "prosemirror-commands";
import { dropCursor } from "prosemirror-dropcursor"
import { gapCursor } from "prosemirror-gapcursor"
import {
  menuBar,
  MenuItem,
  Dropdown,
  wrapItem,
  undoItem as prosemirrorUndoItem,
  redoItem as prosemirrorRedoItem,
  liftItem as prosemirrorLiftItem,
} from "prosemirror-menu";
import { yUndoPluginKey } from "y-prosemirror";
import { getDocName } from "./getDocName";
/**@typedef {import('prosemirror-model').Schema} Schema */

/**
 * @typedef {{
 *   schema: Schema
 *   undo: import('y-prosemirror').undo,
 *   redo: import('y-prosemirror').redo,
 *   getIsManualSend: () => boolean,
 *   onToggleManualSendClick: () => void
 * }} Options
 */

/**
 * @param {Options} options
 */
export default function getProsemirrorSetup(options) {
  // A modified `prosemirror-example-setup/src/index.ts`.

  const menuItemsBase = buildMenuItems(options.schema);
  const menuItemsFinal = [
    [
      new Dropdown(
        [
          new MenuItem({
            run: options.undo,
            label: "Undo typing",
            title: prosemirrorUndoItem.spec.title,
            class: prosemirrorUndoItem.spec.class,
            css:   prosemirrorUndoItem.spec.css,
          }),
          new MenuItem({
            run: options.redo,
            label: "Redo typing",
            title: prosemirrorRedoItem.spec.title,
            class: prosemirrorRedoItem.spec.class,
            css:   prosemirrorRedoItem.spec.css,
          }),
          new MenuItem({
            run: (state) => sendToChat(state),
            label: 'Export as .txt file',
          }),
          new MenuItem({
            run: options.onToggleManualSendClick,
            render(_view) {
              const label = document.createElement('label');

              const checkbox = document.createElement('input');
              checkbox.type = 'checkbox';
              const updateCheckboxState = () => {
                checkbox.checked = !options.getIsManualSend();
              }
              updateCheckboxState();

              // A hack to update the state of the checkbox.
              // `view.update(view.props)` wouldn't work, because it only
              // updates CSS (see https://github.com/ProseMirror/prosemirror-menu/blob/4dc66a6b1ab1d966d2834292f05aab224317beb7/src/menu.ts#L54-L70)
              // + I'm not sure if it's free of side-effects.
              label.addEventListener('mousedown', (e) => {
                setTimeout(() => setTimeout(updateCheckboxState));
              });

              label.append(checkbox, ' Auto-save');

              return label;
            },
            class: 'auto-save-menu-item'
          }),
          new MenuItem({
            run: toggleDark,
            class: 'theme-button',
            title: 'Toggle color theme',
            label: 'Toggle theme'
          }),
        ], {
          label: '☰',
          title: 'Menu',
          class: 'hamburger-menu-item',
        },
      ),
      new Dropdown(
        [
          menuItemsBase.makeParagraph,
          menuItemsBase.makeHead1,
          menuItemsBase.makeHead2,
          menuItemsBase.makeHead3,
          wrapItem(options.schema.nodes.bullet_list, {
            label: "Bullet list",
          }),
          wrapItem(options.schema.nodes.ordered_list, {
            label: "Ordered list",
          }),
          new MenuItem({
            run:    prosemirrorLiftItem.spec.run,
            label:  prosemirrorLiftItem.spec.title,
            select: prosemirrorLiftItem.spec.select,
          }),
          wrapItem(options.schema.nodes.blockquote, {
            label: "Block quote",
          }),
          new MenuItem({
            ...menuItemsBase.insertHorizontalRule.spec,
            label: "Insert line",
          }),
        ], {
          label: 'Paragraph',
        },
      ),
      // menuItems.makeParagraph,
      menuItemsBase.toggleStrong,
      menuItemsBase.toggleEm,
      menuItemsBase.toggleCode,
    ],
  ];

  const plugins = [
    buildInputRules(options.schema),
    keymap(buildKeymap(options.schema, undefined)),
    keymap(baseKeymap),
    dropCursor(),
    gapCursor(),
    menuBar({
      // TODO `floating` doesn't work the way we want it to on iOS. See
      // https://codeberg.org/webxdc/editor/issues/11
      // Also maybe just replace it with `position: sticky`, like here:
      // https://codeberg.org/webxdc/editor/pulls/32
      floating: true,
      // content: menuItems.fullMenu,
      content: menuItemsFinal,
    }),
  ]
  return plugins.concat(new Plugin({
    props: {
      attributes: {class: "ProseMirror-example-setup-style"}
    }
  }))
}

function toggleDark() {
  const agentPrefersDark = window.matchMedia
    ?.('(prefers-color-scheme: dark)').matches
  const newWantedThemeIsDark =
    !document.documentElement.classList.contains('dark');
  if (agentPrefersDark === newWantedThemeIsDark) {
    // System theme and wanted theme match, let's just keep it that way.
    localStorage.removeItem('color-schema');
  } else {
    localStorage.setItem(
      'color-schema',
      newWantedThemeIsDark ? 'dark' : 'light'
    );
  }
  document.documentElement.classList.toggle('dark');
}

/**
 * @param {import("prosemirror-state").EditorState} state
 */
async function sendToChat(state) {
  const serializer = (await import("prosemirror-markdown")).defaultMarkdownSerializer;
  const markdown = serializer.serialize(state.doc);
  await webxdc.sendToChat({
    file: {
      name: 'export.txt',
      plainText: markdown,
    },
    // TODO maybe use this as a file name intead, when we have better idea
    // of how invalid file names are supposed to be handled.
    text: getDocName(state),
  });
}

// Looked up in `y-prosemirror/src/plugins/undo-plugin.js
function getYjsUndoManager(state) {
  const undoManager = /**@type {import('yjs').UndoManager | undefined}*/(
    yUndoPluginKey.getState(state).undoManager
  );
  return undoManager;
}
